#!/usr/bin/env python
# coding: utf-8
#%%
"""
Simple client code for the ESP32 in Python
Copyright 2021 Benedict Diederich, released under LGPL 3.0 or later
"""
from .mserial import Serial
from .mserial import SerialManagerWrapper
from .galvo import Galvo
from .config import config
from .ledmatrix import LedMatrix
from .motor import Motor
from .gripper import Gripper
from .home import Home
from .objective import Objective
from .state import State
from .laser import Laser
from .wifi import Wifi
from .camera import Camera
from .analog import Analog
from .modules import Modules
from .digitalout import DigitalOut
from .rotator import Rotator
from .logger import Logger
from .cmdrecorder import cmdRecorder
from .temperature import Temperature
from .message import Message
try:
    import requests
except:
    print("No requests available - running on pyscript?")

class UC2Client(object):
    # headers = {'ESP32-version': '*'}
    headers={"Content-Type":"application/json"}
    getmessage = ""
    is_connected = False

    is_wifi = False
    is_serial = False

    # BAUDRATE = 500000
    BAUDRATE = 500000

    def __init__(self, host=None, port=31950, serialport=None, identity="UC2_Feather", baudrate=BAUDRATE, NLeds=64, SerialManager=None, DEBUG=False, logger=None, skipFirmwareCheck=False):
        '''
        This client connects to the UC2-REST microcontroller that can be found here
        https://github.com/openUC2/UC2-REST

        generally speaking you send/receive JSON documents that will cause an:
        1. action => "/XXX_act"
        2. getting => "/XXX_get"
        3. setting => "/XXX_set"

        you can send commands through wifi/http or usb/serial
        '''
        if logger is None:
            self.logger = Logger()
        else:
            self.logger = logger

        # perhaps we are in the browser?
        self.isPyScript = False

        # initialize communication channel (# connect to wifi or usb)
        if serialport is not None:
            # use USB connection
            self.serial = Serial(serialport, baudrate, parent=self, identity=identity, DEBUG=DEBUG, skipFirmwareCheck=skipFirmwareCheck)
            self.is_serial = True
            self.is_connected = self.serial.is_connected
            self.serial.DEBUG = DEBUG
        elif host is not None:
            # use client in wireless mode
            self.is_wifi = True
            self.host = host
            self.port = port

            # check if host is up
            self.logger.debug(f"Connecting to microscope {self.host}:{self.port}")
            #self.is_connected = self.isConnected()
        elif SerialManager is not None:
            # we are trying to access the controller from .a web browser
            self.serial = SerialManagerWrapper(SerialManager, parent=self)
            self.isPyScript = True
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")
        
                        
        # import libraries depending on API version
        self.logger.debug("Using API version 2")        

        #FIXME
        #self.set_state(debug=False)

        # initialize state
        self.state = State(self)
        if not self.isPyScript: 
            state = self.state.get_state()
       
        # initialize config
        if not self.isPyScript: 
            self.config = config(self)

        # initialize cmdRecorder
        self.cmdRecorder = cmdRecorder(self)
        
        # initialize LED matrix
        self.led = LedMatrix(self, NLeds=NLeds)

        # initilize motor
        self.motor = Motor(self)
        
        # initialize gripper
        self.gripper = Gripper(self)
        
        # initialize rotator
        self.rotator = Rotator(self)
        
        # initiliaze homing
        self.home = Home(self)
        
        # initialize objective
        self.objective = Objective(self)
        
        # initialize temperature
        self.temperature = Temperature(self)

        # initialize laser
        self.state = State(self)

        # initialize galvo
        self.galvo = Galvo(self)

        # initialize laser
        self.laser = Laser(self)

        # initialize wifi
        self.wifi = Wifi(self)

        # initialize camera
        self.camera = Camera(self)

        # initialize analog
        self.analog = Analog(self)
        
        # initialize digital out
        self.digitalout = DigitalOut(self)
        
        # initialize messaging
        self.message = Message(self)
        
        # initialize config
        if False: # not self.isPyScript: 
            self.config = config(self)
            try: self.pinConfig = self.config.loadConfigDevice()
            except: self.pinConfig = None
        

        # initialize module controller
        self.modules = Modules(parent=self)
    
    def post_json(self, path, payload, getReturn=True, nResponses=1, timeout=1):
        if self.is_wifi:
            # FIXME: this is not working
            url = f"http://{self.host}:{self.port}{path}"
            try:
                if timeout==0: timeout=.2
                r = requests.post(url, json=payload, headers=self.headers,  timeout=timeout)
                returnMessage = r.json()
                returnMessage["success"] = r.status_code==200
            except Exception as e:
                print(e)
                returnMessage = {}
                returnMessage["error"] = str(e)
                returnMessage["success"] = 0
            return returnMessage
        elif self.is_serial or self.isPyScript:
            if timeout <=0:
                getReturn = False
            return self.serial.post_json(path, payload, getReturn=getReturn, nResponses=nResponses, timeout=timeout)
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")
            return None

    def get_json(self, path, getReturn=True, timeout=1):
        if self.is_wifi:
            # FIXME: this is not working
            url = f"http://{self.host}:{self.port}{path}"
            r = requests.get(url, headers=self.headers, timeout=timeout)
            return r.json()
        elif self.is_serial or self.isPyScript:
            # timeout is not used anymore
            if timeout <=0:
                getReturn = False
            return self.serial.post_json(path, payload=None, getReturn=getReturn, nResponses=1, timeout=timeout)
            #return self.serial.read_json()<
        else:
            self.logger.error("No ESP32 device is connected - check IP or Serial port!")
            return None

    def setDebugging(self, debug=False):
        self.logger.debug(f"Setting debugging to {debug}")
        self.serial.DEBUG = debug
        
    def close(self):
        self.serial.close()


import json
class cmdRecorder(object):
    '''
    A class for recording and playing back commands sent to a device.

    Attributes:
    -----------
    _parent : object
        The parent object that the cmdRecorder is attached to.
    commandQueue : list
        A list of commands that have been recorded.

    Methods:
    --------
    startRecording()
        Starts recording commands by storing them in the commandQueue.
    stopRecording()
        Stops recording commands and clears the commandQueue.
    addCommand(command)
        Adds an individual hardware command to the commandQueue.
    printCommands()
        Prints the previously recorded commands.
    playCommands(nTimes=1, isBlocking=True)
        Plays the commands in the commandQueue nTimes.
    '''

    def __init__(self, parent):
        '''
        Initializes a new instance of the cmdRecorder class.

        Parameters:
        -----------
        parent : object
            The parent object that the cmdRecorder is attached to.
        '''
        self._parent = parent
        self.commandQueue = []
        
    def startRecording(self):
        '''
        Starts recording commands by storing them in the commandQueue.
        '''
        self.commandQueue = []
        self._parent.serial.toggleCommandOutput(cmdCallBackFct = self.addCommand)
        
    def stopRecording(self):
        '''
        Stops recording commands and clears the commandQueue.
        '''
        self._parent.serial.toggleCommandOutput(cmdCallBackFct = None)

    def addCommand(self, command):
        '''
        Adds an individual hardware command to the commandQueue.

        Parameters:
        -----------
        command : str
            The command to add to the commandQueue.
        '''
        self.commandQueue.append(command)
        
    def printCommands(self):
        '''
        Prints the previously recorded commands.
        '''
        print(self.commandQueue)
        
    def playCommands(self, nTimes=1, isBlocking=False):
        '''
        Plays the commands in the commandQueue nTimes.
        # like so: 
        # {"tasks":[{"task":"/state_get"},{"task":"/state_act", "delay":1000}],"nTimes":2}
        # {'tasks:': '[{"task": "/state_get"}, {"task": "/state_act", "delay": 1000}]', 'nTimes': 3, 'qid': 2}
        Parameters:
        -----------
        nTimes : int, optional
            The number of times to play the commands. Default is 1.
        isBlocking : bool, optional
            Whether or not to wait for the device to respond before sending the next command. Default is True.
        '''
        payload = {"tasks":self.commandQueue, "nTimes":nTimes}
        nResponses = len(self.commandQueue) * nTimes * isBlocking # not quite true since motors will have multiple returns..but fine for now
        r = self._parent.serial.sendMessage(command=payload, nResponses=nResponses)
        return r


class Laser(object):
    ## Laser
    def __init__(self, parent):
        self._parent = parent
        self.filter_pos_1 = 0
        self.filter_pos_2 = 0
        self.filter_pos_3 = 0
        self.filter_pos_LED = 0
        
    def set_laser(self, channel=1, value=0, auto_filterswitch=False,
                        filter_axis=-1, filter_position = None,
                        despeckleAmplitude = 0.,
                        despecklePeriod=10, timeout=20, is_blocking = False):
        if channel not in (0,1,2,3):
            if channel=="R":
                channel = 1
            elif channel=="G":
                channel = 2
            elif channel=="B":
                channel = 3

        if auto_filterswitch and value >0:
            if filter_position is None:
                if channel==1:
                    filter_position_toGo = self.filter_pos_1
                if channel==2:
                    filter_position_toGo = self.filter_pos_2
                if channel==3:
                    filter_position_toGo = self.filter_pos_3
                if channel=="LED":
                    filter_position_toGo = self.filter_pos_LED
            else:
                filter_position_toGo = filter_position

            self.switch_filter(filter_pos=filter_position_toGo, filter_axis=filter_axis, timeout=timeout,is_blocking=is_blocking)

        path = '/laser_act'
        
        payload = {
            "task": path,
            "LASERid": channel,
            "LASERval": value,
            "LASERdespeckle": int(value*despeckleAmplitude),
            "LASERdespecklePeriod": int(despecklePeriod),

        }
        #self._parent.logger.debug("Setting Laser "+str(channel)+", value: "+str(value))
        r = self._parent.post_json(path, payload, getReturn=is_blocking, timeout=.5)
        return r

    def set_laserpin(self, laserid=1, laserpin=0):
        path = '/laser_set'
        
        payload = {
            "task": path,
            "LASERid": laserid,
            "LASERpin": laserpin
        }
        
        r = self._parent.post_json(path, payload)
        return r



    def set_servo(self, channel=1, value=0, is_blocking=False):
        #{"task":"/laser_act", "LASERid":1 ,"LASERval":99, "servo":1, "qid":1}
        path = '/laser_act'
        
        payload = {
            "task": path,
            "LASERid": channel,
            "LASERval": value, 
            "servo": 1
        }
        
        r = self._parent.post_json(path, payload, getReturn=is_blocking)
        return r


import numpy as np
import time
import json


gTIMEOUT = 100 # seconds to wait for a response from the ESP32
class Motor(object):

    # indicate if there is any motion happening
    isRunning = False

    def __init__(self, parent=None):
        self._parent = parent

        # do we have a coreXY setup?
        self.isCoreXY = False

        self.nMotors = 4
        self.lastDirection = np.zeros((self.nMotors))
        self.backlash = np.zeros((self.nMotors))
        self.stepSize = np.ones((self.nMotors))
        self.maxStep = np.ones((self.nMotors))*np.inf
        self.minStep = np.ones((self.nMotors))*(-np.inf)
        self.currentDirection = np.zeros((self.nMotors))
        self.currentPosition = np.zeros((self.nMotors))
        self._position = np.zeros((self.nMotors)) # position from the last motor status update

        self.minPosX = -np.inf
        self.minPosY = -np.inf
        self.minPosZ = -np.inf
        self.minPosA = -np.inf
        self.maxPosX = np.inf
        self.maxPosY = np.inf
        self.maxPosZ = np.inf
        self.maxPosA = np.inf
        self.stepSizeX =  1
        self.stepSizeY =  1
        self.stepSizeZ =  1
        self.stepSizeA =  1
        self.offsetA = 0
        self.offsetX = 0
        self.offsetY = 0
        self.offsetZ = 0

        self.DEFAULT_ACCELERATION = 10000

        self.motorAxisOrder = [0,1,2,3] # motor axis is 1,2,3,0 => X,Y,Z,T # FIXME: Hardcoded

        # register a callback function for the motor status on the serial loop
        if hasattr(self._parent, "serial"):
            self._parent.serial.register_callback(self._callback_motor_status, pattern="steppers")
        # announce a function that is called when we receive a position update through the callback
        self._callbackPerKey = {}
        self.nCallbacks = 10
        self._callbackPerKey = self.init_callback_functions(nCallbacks=self.nCallbacks) # only one is used for now
        print(self._callbackPerKey)
        # move motor to wake them up #FIXME: Should not be necessary!
        #self.move_stepper(steps=(1,1,1,1), speed=(1000,1000,1000,1000), is_absolute=(False,False,False,False))
        #self.move_stepper(steps=(-1,-1,-1,-1), speed=(1000,1000,1000,1000), is_absolute=(False,False,False,False))

    def init_callback_functions(self, nCallbacks=10):
        ''' initialize the callback functions '''
        _callbackPerKey = {}
        self.nCallbacks = nCallbacks
        for i in range(nCallbacks):
            _callbackPerKey[i] = []
        return _callbackPerKey
            
    def _callback_motor_status(self, data):
        ''' cast the json in the form:
        {
        "qid":	0,
        "steppers":	[{
                "stepperid":	1,
                "position":	1000,
                "isDone":	1
            }]
        }
        into the position array of the motors '''
        try:
            nSteppers = len(data["steppers"])
            stepSizes = np.array((self.stepSizeA, self.stepSizeX, self.stepSizeY, self.stepSizeZ))
            offSets = np.array((self.offsetA, self.offsetX, self.offsetY, self.offsetZ))
            for iMotor in range(nSteppers):
                stepperID = data["steppers"][iMotor]["stepperid"]
                # FIXME:  smart to re-update this variable? Will be updated by motor-sender too
                self._position[stepperID] = data["steppers"][iMotor]["position"] * stepSizes[stepperID] - offSets[stepperID]
            if  callable(self._callbackPerKey[0]):
                self._callbackPerKey[0](self._position) # we call the function with the value
        except Exception as e:
            print("Error in _callback_motor_status: ", e)

    def register_callback(self, key, callbackfct):
        ''' register a callback function for a specific key '''
        self._callbackPerKey[key] = callbackfct
        
        
    def setTrigger(self, axis="X", pin=1, offset=0, period=1):
        # {"task": "/motor_act", "setTrig": {"steppers": [{"stepperid": 1, "trigPin": 1, "trigOff":0, "trigPer":1}]}}
        if type(axis) is not int:
            axis = self.xyztTo1230(axis)
        path = "/motor_act"
        payload = {
            "task": path,
            "setTrig":{
                "steppers": [
                {
                 "stepperid": axis,
                 "trigPin": pin,
                 "trigOff": offset,
                 "trigPer": period
                 }]
            }}
        r = self._parent.post_json(path, payload)
        return r

    # {"task": "/motor_act", "stagescan": {"nStepsLine": 50, "dStepsLine": 1, "nTriggerLine": 1, "nStepsPixel": 50, "dStepsPixel": 1, "nTriggerPixel": 1, "delayTimeStep": 10, "stopped": 0, "nFrames": 50}}"}}
    def startStageScanning(self, nStepsLine=100, dStepsLine=1, nTriggerLine=1, nStepsPixel=100, dStepsPixel=1, nTriggerPixel=1, delayTimeStep=10, nFrames=5, isBlocking = False):
        path = "/motor_act"
        payload = {
            "task": path,
            "stagescan":{
                "nStepsLine": nStepsLine,
                "dStepsLine": dStepsLine,
                "nTriggerLine": nTriggerLine,
                "nStepsPixel": nStepsPixel,
                "dStepsPixel": dStepsPixel,
                "nTriggerPixel": nTriggerPixel,
                "delayTimeStep": delayTimeStep,
                "stopped": 0,
                "nFrames": nFrames
            }}
        r = self._parent.post_json(path, payload, getReturn=isBlocking)
        return r

    def stopStageScanning(self):
        self.startStageScanning(stopped=1)

    def setIsCoreXY(self, isCoreXY = False):
        self.isCoreXY = isCoreXY

    def setMotorAxisOrder(self, order=[0,1,2,3]):
        self.motorAxisOrder = order

    '''################################################################################################################################################
    HIGH-LEVEL Functions that rely on basic REST-API functions
    ################################################################################################################################################'''
    def setup_motor(self, axis, minPos, maxPos, stepSize, backlash, offset=0):
        if axis == "X":
            self.minPosX = minPos
            self.maxPosX = maxPos
            self.stepSizeX = stepSize
            self.offsetX = offset
        elif axis == "Y":
            self.minPosY = minPos
            self.maxPosY = maxPos
            self.stepSizeY = stepSize
            self.offsetY = offset
        elif axis == "Z":
            self.minPosZ = minPos
            self.maxPosZ = maxPos
            self.stepSizeZ = stepSize
            self.offsetZ = offset
        elif axis == "A":
            self.minPosA = minPos
            self.maxPosA = maxPos
            self.stepSizeA = stepSize
            self.offsetA = offset   
        self.backlash[self.xyztTo1230(axis)] = backlash

    def xyztTo1230(self, axis):
        axis = axis.upper()
        if axis == "X":
            axis = 1
        if axis == "Y":
            axis = 2
        if axis == "Z":
            axis = 3
        if axis == "A":
            axis = 0
        return axis

    def cartesian2corexy(self, x, y):
        # convert cartesian coordinates to coreXY coordinates
        # https://www.corexy.com/theory.html
        x1 = (x+y)/np.sqrt(2)
        y1 = (x-y)/np.sqrt(2)
        return x1, y1

    def move_x(self, steps=0, speed=1000, acceleration=None, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        if self.isCoreXY:
            # have to turn two motors to move in X direction
            xTemp, yTemp =  self.cartesian2corexy(steps, 0)
            return self.move_xy(steps=(xTemp, yTemp), speed=(speed,speed), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)
        else:
            return self.move_axis_by_name(axis="X", steps=steps, speed=speed, acceleration=acceleration, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)

    def move_y(self, steps=0, speed=1000, acceleration=None,  is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        if self.isCoreXY:
            # have to turn two motors to move in Y direction
            xTemp, yTemp =  self.cartesian2corexy(0,steps)
            return self.move_xy(steps=(xTemp, yTemp), speed=(speed,speed), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)
        else:
            return self.move_axis_by_name(axis="Y", steps=steps, speed=speed, acceleration=acceleration, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)

    def move_z(self, steps=0, speed=1000, acceleration=None,  is_blocking=False, is_absolute=False, is_dualaxis = False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        if is_dualaxis:
            self.move_az(steps=(steps, steps), speed=(speed,speed), acceleration=acceleration, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=gTIMEOUT, is_reduced=is_reduced)
        else:
            return self.move_axis_by_name(axis="Z", steps=steps, speed=speed, acceleration=acceleration, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)

    def move_a(self, steps=0, speed=1000, acceleration=None, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        return self.move_axis_by_name(axis="A", steps=steps, speed=speed, acceleration=acceleration, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)

    def move_xyz(self, steps=(0,0,0), speed=(1000,1000,1000), acceleration=None, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        if len(speed)!= 3:
            speed = (speed,speed,speed)

        # motor axis is 1,2,3,0 => X,Y,Z,T # FIXME: Hardcoded
        r = self.move_xyza(steps=(0,steps[0],steps[1],steps[2]), acceleration=(0,acceleration[0],acceleration[1],acceleration[2]), speed=(0,speed[0],speed[1],speed[2]), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)
        return r

    def move_xy(self, steps=(0,0), speed=(1000,1000), acceleration=None, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        if self.isCoreXY:
            # have to move only one motor to move in XY direction
           return self.move_xyza(steps=(0,steps[0], steps[1], 0), speed=(0,speed[0],speed[1],0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)
        else:
            if type(speed)==int or len(speed)!= 2:
                speed = (speed,speed)

            if acceleration is None:
                acceleration = 100000
                
            if type(acceleration)==int or len(acceleration)!= 2:
                acceleration = (acceleration,acceleration)

            # motor axis is 1,2,3,0 => X,Y,Z,T # FIXME: Hardcoded
            r = self.move_xyza(steps=(0, steps[0],steps[1],0), speed=(0,speed[0],speed[1],0), acceleration=(0,acceleration[0],acceleration[1],0), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)
            return r

    def move_az(self, steps=(0,0), speed=(1000,1000), acceleration=None, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        if (type(speed)!=list and type(speed)!=tuple) or len(speed)!= 2:
            speed = (speed,speed)

        if (type(acceleration)!=list and type(acceleration)!=tuple) or len(acceleration)!= 2:
            acceleration = (acceleration,acceleration)

        # motor axis is 1,2,3,0 => X,Y,Z,T # FIXME: Hardcoded
        r = self.move_xyza(steps=(steps[0],0,0,steps[1]), speed=(speed[0],0,0,speed[1]), acceleration=(acceleration[0],0,0,acceleration[1]), is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)
        return r

    def move_xyza(self, steps=(0,0,0,0), speed=(1000,1000,1000,1000), acceleration=None, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        # everywhere, where relative steps are zero, speed should be zero, too
        if type(speed)==int:
            speed = [speed,speed,speed,speed]
        if type(speed)==tuple:
            speed = list(speed)
        for iMotor in range(len(steps)):
            if steps[iMotor]==0 and not is_absolute:
                speed[iMotor] = 0

        r = self.move_stepper(steps=steps, speed=speed, acceleration=acceleration, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, timeout=timeout, is_reduced=is_reduced)
        return r

    def move_axis_by_name(self, axis="X", steps=100, speed=1000, acceleration=None, is_blocking=False, is_absolute=False, is_enabled=True, timeout=gTIMEOUT, is_reduced=False):
        axis = self.xyztTo1230(axis)
        _speed=np.zeros(4)
        _speed[axis] = speed
        _steps=np.array((0,0,0,0))
        _steps[axis] = steps
        _acceleration=acceleration
        r = self.move_stepper(_steps, speed=_speed, acceleration=_acceleration, timeout=timeout, is_blocking=is_blocking, is_absolute=is_absolute, is_enabled=is_enabled, is_reduced=is_reduced)
        return r

    def move_forever(self, speed=(0,0,0,0), is_stop=False, is_blocking=False):
        if type(speed)==int:
            speed=(speed, speed, speed, speed)
        if len(speed)==3:
            speed = (*speed,0)
        '''
        {"task":"/motor_act",
            "motor":
            {
                "steppers": [
                    { "stepperid": 3, "isforever": 1, "speed": 2000}
                ]
            }
        }
        '''
        # only consider those actions that are necessary
        motorPropList = []
        for iMotor in range(4):
            if abs(speed[iMotor])>0:
                motorProp = { "stepperid": iMotor,
                             "isforever": int(not is_stop),
                             "speed": speed[iMotor]}
                motorPropList.append(motorProp)

        path = "/motor_act"
        payload = {
            "task":path,
            "motor":
            {
                "steppers": motorPropList
            }
        }

        r = self._parent.post_json(path, payload, timeout=0, getReturn=is_blocking)
        return r

    def stop(self, axis=None):
        axisNumberList = []
        if axis is None:
            for iMotor in range(self.nMotors):
                axisNumberList.append(iMotor)
        else:
            axisNumberList.append(self.xyztTo1230(axis))

        motorPropList = []
        for iMotor in axisNumberList:
            motorProp = { "stepperid": self.motorAxisOrder[iMotor],
                            "isstop": True}
            motorPropList.append(motorProp)

        path = "/motor_act"
        payload = {
            "task":path,
            "motor":
            {
                "steppers": motorPropList
            }
        }
        # ensure that nothing blocks this command!
        if self.isRunning:
            self._parent.serial.breakCurrentCommunication()
        r = self._parent.post_json(path, payload, getReturn=False, timeout=0)
        return r

    def compute_travel_time(self, steps, max_speed, acceleration):
        import math
        """
        Calculate approximate travel time for a stepper motor 
        with trapezoidal velocity profile.

        steps       : total distance in steps
        max_speed   : maximum speed in steps/second
        acceleration: acceleration in steps/second^2
        """
        # Distance to accelerate from 0 to max_speed:
        dist_accel = (max_speed ** 2) / (2.0 * acceleration)

        # If distance is too short to reach max speed -> triangular profile
        if steps < 2 * dist_accel:
            # T = 2 * sqrt( distance / acceleration )
            return 2.0 * math.sqrt(steps / acceleration)
        else:
            # Time to accelerate to max_speed
            t_accel = max_speed / acceleration
            # Distance spent accelerating
            d_accel = dist_accel
            # Distance spent decelerating (same as accelerating)
            d_decel = dist_accel
            # Remaining distance at constant speed
            d_const = steps - d_accel - d_decel
            # Time at constant speed
            t_const = d_const / max_speed
            # Total time
            return t_accel + t_const + t_accel



    def move_stepper(self, steps=(0,0,0,0), speed=(1000,1000,1000,1000), is_absolute=(False,False,False,False), timeout=gTIMEOUT, acceleration=(None, None, None, None), is_blocking=True, is_enabled=True, is_reduced=False):
        '''
        This tells the motor to run at a given speed for a specific number of steps; Multiple motors can run simultaneously

        XYZT => 1,2,3,0
        '''
        # determine the axis to operate
        axisToMove = np.where(np.abs(speed)>0)[0]
        if axisToMove.shape[0]==0:
            return "{'return':-1}"
        if type(is_absolute)==bool:
            isAbsoluteArray = np.zeros((4))
            isAbsoluteArray[axisToMove] = is_absolute
        else:
            isAbsoluteArray = is_absolute

        # convert single elements to array
        if type(speed)!=list and type(speed)!=tuple and type(speed)!=np.ndarray:
            speed = np.array((speed,speed,speed,speed))

        # convert single elements to array
        if type(acceleration)!=list and type(acceleration)!=tuple and type(acceleration)!=np.ndarray:
            acceleration = np.array((acceleration,acceleration,acceleration,acceleration))

        # make sure value is an array
        if type(steps)==tuple:
            steps = np.array(steps)

        # optionally add the offset to the steps
        if isAbsoluteArray[0]: 
            steps[0] += self.offsetA
        if isAbsoluteArray[1]:
            steps[1] += self.offsetX
        if isAbsoluteArray[2]:
            steps[2] += self.offsetY
        if isAbsoluteArray[3]:
            steps[3] += self.offsetZ

        # convert from physical units to steps
        steps[0] *= 1/self.stepSizeA
        steps[1] *= 1/self.stepSizeX
        steps[2] *= 1/self.stepSizeY
        steps[3] *= 1/self.stepSizeZ
        
        # detect change in direction
        absoluteDistances = np.zeros((4))
        for iMotor in range(4):
            # for absolute motion:
            if isAbsoluteArray[iMotor]:
                self.currentDirection[iMotor] = 1 if (self.currentPosition[iMotor]  > steps[iMotor]) else -1
            else:
                self.currentDirection[iMotor] = np.sign(steps[iMotor])
            if self.lastDirection[iMotor] != self.currentDirection[iMotor]:
                # we want to overshoot a bit
                steps[iMotor] = steps[iMotor] +  self.currentDirection[iMotor]*self.backlash[iMotor]

    
            if isAbsoluteArray[iMotor]:
                absoluteDistances[iMotor] = abs(self.currentPosition[iMotor] - steps[iMotor])
            else:
                absoluteDistances[iMotor] = abs(steps[iMotor])
        # experimental: Let's make the timeout adaptive: 
        # the speed of the motors is given in steps/second, so we can calculate the time it takes to move the given steps
        # we will add a bit of time to the timeout to make sure we get a return
        # 1.5 accounts for accel/decceleration
        traveltime = self.compute_travel_time(np.max(np.abs(absoluteDistances)), np.max(np.abs(speed)), np.max(np.where(acceleration == None, 20000, acceleration)))
        timeout = np.uint8(abs(timeout)>0)*(traveltime + 1) # add 1 second to the timeout to make sure we get a return

        # get current position
        #_positions = self.get_position() # x,y,z,t = 1,2,3,0
        #pos_3, pos_0, pos_1, pos_2 = _positions[0],_positions[1],_positions[2],_positions[3]


        '''
        # check if within limits
        if pos_0+steps_0 > self.maxPosX or pos_0+steps_0 < self.minPosX:
            steps_0=0
        if pos_1+steps_1 > self.maxPosY or pos_1+steps_1 < self.minPosY:
            steps_1=0
        if pos_2+steps_2 > self.maxPosZ or pos_2+steps_2 < self.minPosZ:
            steps_2 = 0
        if pos_3+steps_3 > self.maxPosA or pos_3+steps_3 < self.minPosA:
            steps_3 = 0
        '''


        # only consider those actions that are necessary
        motorPropList = []
        for iMotor in range(4):
            if isAbsoluteArray[iMotor] or abs(steps[iMotor])>0:
                motorProp = { "stepperid": int(self.motorAxisOrder[iMotor]),
                             "position": int(steps[iMotor]),
                             "speed": int(speed[iMotor]),
                             "isabs": int(isAbsoluteArray[iMotor]),
                             "isaccel": int(1),
                             "isen": int(is_enabled), 
                             "redu": int(is_reduced)}
                if acceleration[iMotor] is not None:
                    motorProp["accel"] = int(acceleration[iMotor])
                else:
                    motorProp["accel"] = self.DEFAULT_ACCELERATION
                motorPropList.append(motorProp)

        path = "/motor_act"
        payload = {
            "task":path,
            "motor":
            {
                "steppers": motorPropList
            }
        }

        # safe steps to track direction for backlash compensatio
        for iMotor in range(self.nMotors):
            if isAbsoluteArray[iMotor]:
                self.currentPosition[iMotor] = steps[iMotor]
            else:
                self.currentPosition[iMotor] = self.currentPosition[iMotor] + steps[iMotor]

        # drive motor
        self.isRunning = True
        is_blocking = not self._parent.is_wifi and is_blocking and self._parent.serial.is_connected
        timeout = timeout if is_blocking else 0
        if type(axisToMove) == list:
            nResponses = len(axisToMove)+1 # we get the command received flag + a return for every axis
        elif type(axisToMove) == tuple:
            nResponses = axisToMove[0].shape[0]+1
        elif type(axisToMove) == np.ndarray:
            nResponses = axisToMove.shape[0] +1
        else:
            nResponses = 2
        # if we get a return, we will receive the latest position feedback from the driver  by means of the axis that moves the longest
        r = self._parent.post_json(path, payload, getReturn=is_blocking, timeout=timeout, nResponses=nResponses)


        # save direction for last iteration
        self.lastDirection = self.currentDirection.copy()

        # Reset busy flag
        self.isRunning = False

        return r

    def isBusy(self, steps, timeout=1):
        path = "/motor_get"
        payload = {
            "task":path,
            "isbusy": 1
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        try:
            isbusy = 0
            for iMotor in range(self.nMotors):
                isbusy += r["motor"]["steppers"][iMotor]["isbusy"]
            if isbusy:
                return True
            else :
                return False
        except Exception as e:
            return False



    def set_motor(self, stepperid = 0, position = None, stepPin=None, dirPin=None, enablePin=None, maxPos=None, minPos=None, acceleration=None, isEnable=None, isBlocking=True, timeout=2):
        path = "/motor_set"
        payload = {"task":path,
                    "motor":{
                    "steppers": [
                        {   "stepperid": self.motorAxisOrder[stepperid]}]
                    }}

        if stepPin is not None: payload['motor']["steppers"][0]["step"] = stepPin
        if dirPin is not None: payload['motor']["steppers"][0]["dir"] = dirPin
        if enablePin is not None: payload['motor']["steppers"][0]["enable"] = enablePin
        if maxPos is not None: payload['motor']["steppers"][0]["max_pos"] = maxPos
        if minPos is not None: payload['motor']["steppers"][0]["min_pos"] = minPos
        if position is not None: payload['motor']["steppers"][0]["position"] = position
        if acceleration is not None: payload['motor']["steppers"][0]["acceleration"] = acceleration
        if isEnable is not None: payload['motor']["steppers"][0]["isen"] = isEnable

        # send command
        r = self._parent.post_json(path, payload, timeout=1)

        # wait until job has been done
        time0=time.time()
        if isBlocking:
            while self.isBusy((0,0,0,0)):
                time.sleep(0.1)
                if time.time()-time0>timeout:
                    break

        return r

    def set_motor_currentPosition(self, axis=0, currentPosition=10000):
        if type(axis)==str:
            axis = self.xyztTo1230(axis)

        r = self.set_motor(stepperid = axis, position = currentPosition)
        return r

    def set_motor_acceleration(self, axis=0, acceleration=40000):
        if type(axis)==str:
            axis = self.xyztTo1230(axis)

        path = "/motor_act"
        payload = {
            "task": path,
            "motor":
            {
            "steppers": [
            { "stepperid":axis, "isaccel":1, "accel":acceleration}
            ]
            }
            }
        r = self._parent.post_json(path, payload)
        return r


    def set_motor_enable(self, enable=None, enableauto=None):
        """
        turns on/off enable pin overrides motor settings - god for cooling puproses
        eanbaleauto  turns on/off timer of the accelstepper library
        """
        path = "/motor_act"
        payload = {
            "task": path
        }
        if enable is not None:
            payload["isen"] = enable
        if enableauto is not None:
            payload["isenauto"] = enableauto
        if not enableauto:
            enable = True
        r = self._parent.post_json(path, payload)
        return r

    def get_position(self, axis=None, timeout=1):
        # pulls all current positions from the stepper controller
        path = "/motor_get"
        payload = {
            "task":path,
            "position":True,
        }
        _position = np.array((0.,0.,0.,0.)) # T,X,Y,Z
        _physicalStepSizes = np.array((self.stepSizeA, self.stepSizeX, self.stepSizeY, self.stepSizeZ))
        _physicalOffsets = np.array((self.offsetA, self.offsetX, self.offsetY, self.offsetZ))

        # this may be an asynchronous call.. #FIXME!
        r = self._parent.post_json(path, payload, getReturn = True, nResponses=1, timeout=timeout)
        # returns {"motor": }
        if "motor" in r:
            for index, istepper in enumerate(r["motor"]["steppers"]):
                if index >3: break # TODO: We would need to handle other values too soon
                _position[istepper["stepperid"]]=istepper["position"]*_physicalStepSizes[self.motorAxisOrder[index]]-_physicalOffsets[self.motorAxisOrder[index]]


        return _position

    def set_position(self, axis=1, position=0, timeout=1):

        '''
        {"task":"/motor_act",  "setpos": {"steppers": [{"stepperid":1, "posval": 0}]}}
        '''
        path = "/motor_act"
        if type(axis) !=int:
            axis = self.xyztTo1230(axis)

        payload = {
            "task": path,
            "setpos":{
                "steppers": [
                {
                 "stepperid": axis,
                 "posval": int(position)
                 }]
            }}
        r = self._parent.post_json(path, payload, timeout=timeout)

        return r

    def set_offset(self, axis=1, offset=0):
        '''
        This sets the offset relative to the homed position, helpful if you have a reference point and you need to compute the offset
        '''
        if axis == 0 or str(axis).upper() == "A":
            self.offsetA = offset
        elif axis == 1 or str(axis) == "X":
            self.offsetX = offset
        elif axis == 2 or str(axis) == "Y":
            self.offsetY = offset
        elif axis == 3 or str(axis) == "Z":
            self.offsetZ = offset
        else:
            self._parent.logger.error("Axis not recognized")
            
    def get_offset(self, axis=1):   
        '''
        This gets the offset relative to the homed position, helpful if you have a reference point and you need to compute the offset
        '''
        if axis == 0 or axis == "A":
            return self.offsetA
        elif axis == 1 or axis == "X":
            return self.offsetX
        elif axis == 2 or axis == "Y":
            return self.offsetY
        elif axis == 3 or axis == "Z":
            return self.offsetZ
        else:
            return False
        
    def get_motor(self, axis=1, timeout=1):
        path = "/motor_get"
        payload = {
            "task":path,
        }
        r = self._parent.post_json(path, payload, timeout=timeout)
        try: return r["steppers"][self.motorAxisOrder[axis]]
        except: return False

    def get_motors(self, timeout=1):
        path = "/motor_get"

        r = self._parent.get_json(path, timeout=timeout)
        try: return r["steppers"]
        except: return False


    def set_direction(self, axis=1, sign=1, timeout=1):
        return False

    def set_tmc_parameters(self, axis=0, msteps=None, rms_current=None, stall_value=None, sgthrs=None, semin=None, semax=None, blank_time=None, toff=None, timeout=1):
        ''' set the TMC parameters for a specific axis 
        msteps: microsteps
        rms_current: current in mA
        sgthrs: stallguard threshold
        semin: minimum current
        semax: maximum current
        blank_time: blank time
        toff: off time
        #   {"task":"/tmc_act", "msteps":32, "rms_current":600, "sgthrs":100, "semin":5, "semax":2, "blank_time":24, "toff":4, "axis":1}
        '''
        if type(axis)==str:
            axis = self.xyztTo1230(axis)
        path = "/tmc_act"
        payload = {}
        if axis is not None:
            payload["axis"] = axis
        if msteps is not None:
            payload["msteps"] = msteps
        if rms_current is not None:
            payload["rms_current"] = rms_current
        if sgthrs is not None:
            payload["sgthrs"] = sgthrs
        if semin is not None:
            payload["semin"] = semin
        if semax is not None:
            payload["semax"] = semax
        if blank_time is not None:
            payload["blank_time"] = blank_time
        if toff is not None:
            payload["toff"] = toff

        r = self._parent.post_json(path, payload, timeout=timeout)
        return r


import serial
import json
import queue
import threading
import time
import sys
try:
    from serial.tools  import list_ports
    from serial.serialutil import SerialException
    IS_SERIAL = True
except ImportError:
    IS_SERIAL = False
    class SerialException(Exception):
        pass

T_SERIAL_WARMUP = 2.5
class Serial:
    def __init__(self, port, baudrate=115200, timeout=5,
                 identity="UC2_Feather", parent=None, DEBUG=False, 
                 skipFirmwareCheck=False):

        self.serialdevice = None
        self.serialport = port
        self.baudrate = baudrate
        self.timeout = timeout
        self._parent = parent
        self.manufacturer = ""
        self.skipFirmwareCheck = skipFirmwareCheck
        if self._parent is None:
            import logging
            self._logger = logging.getLogger(__name__)
            self._logger.setLevel(logging.DEBUG)
            self._logger.addHandler(logging.StreamHandler())
        else:
            self._logger = self._parent.logger

        self.identity = identity
        self.DEBUG = DEBUG
        self.is_connected = False
        self.write_timeout = 1
        self.read_timeout = 0.02
        self.timeReturnReceived = 1#  0.5

        self.cmdWriteCallBackFct = None
        self.cmdReadCallBackFct = None

        # setup command queue
        self.resetLastCommand = False
        self.sender_queue = queue.Queue()
        self.responses = {}
        self.commands = {}
        self.lock = threading.Lock()
        self.serialLock = threading.Lock()
        
        # setup callback list for parent modules
        self.callBackList = []

        # initialize serial connection
        self.thread = None
        if IS_SERIAL:
            self.serialdevice= self.openDevice(port, baudrate)
        else: 
            ''' this is most likely happening because we work under pyoidide '''
            self._logger.debug("You have to provide the serial interface on your own")

    def breakCurrentCommunication(self):
        self.resetLastCommand = True

    def _freeSerialBuffer(self, ser, timeout=5, timeMinimum=0):
        t0 = time.time()
        # free up any old data
        while True:
            try:
                readLineRaw = self._read(ser)   
                readLine = readLineRaw.decode('utf-8').strip()
                if self.DEBUG and readLine != "": 
                    self._logger.debug(readLine)
                if readLine == "" and time.time()-t0 > timeMinimum:
                    break
            except Exception as e:
                if self.DEBUG: self._logger.error("[FreeSerial]: "+str(e))

                pass
            if time.time()-t0 > timeout:
                return


    def get_port_info(self, port_name):
        """
        Überprüft, ob ein Portname in der Liste der verfügbaren Ports enthalten ist,
        und gibt die zugehörigen Portinformationen zurück.

        :param port_name: Der Name des Ports (z. B. '/dev/ttyUSB0')
        :return: Portinformationen oder None, wenn der Port nicht gefunden wurde
        """
        available_ports = list_ports.comports()
        for port in available_ports:
            if port.device == port_name:
                return port  # Gibt die Portinformationen zurück
        return None  # Port nicht gefunden


    def openDevice(self, port=None, baud_rate=115200):
        try: # try to close an eventually open serial connection
            if str(type(self.ser)) != "<class 'uc2rest.mserial.MockSerial'>":
                self.serialdevice.close()
        except: pass

        
        try:
            # check if port is string and if so within available ports 
            if type(port)==str:
                port = self.get_port_info(port)
                if port is None:
                    raise ValueError("Port not found")
                
            for i in range(2): # not good, but sometimes it  needs a second attempt
                isUC2 = self.tryToConnect(port)
                if isUC2:
                    break
            if not isUC2:
                raise ValueError('Wrong Firmware.')
            ser = self.serialdevice
            self.is_connected = True

        except Exception as e:
            self._logger.error("[OpenDevice]: "+str(e))
            ser = self.findCorrectSerialDevice()
            if ser is None:
                ser = MockSerial(port, baud_rate, timeout=.1)
                self.is_connected = False
        ser.write_timeout = self.write_timeout
        if not ser.isOpen():
            ser.open()
        # TODO: Need to be able to auto-connect
        # need to let device warm up and flush out any old data
        self._freeSerialBuffer(ser)

        # remove any remaining thread in case there was one open
        try:
            del self.thread
        except:
            pass
        self.running = True
        self.identifier_counter = 0 # Counter for generating unique identifiers
        self.thread = threading.Thread(target=self._process_commands, daemon=True)
        self.thread.start()
        
        # setup sender queue
        self.sender_worker_thread = threading.Thread(target=self._sending_commands)
        self.sender_worker_thread.daemon = True  # Ensure the thread exits when the main program does
        self.sender_worker_thread.start()  
        return ser

    def findCorrectSerialDevice(self):
        '''
        This function tries to find the correct serial device from the list of available ports
        It also checks if the firmware is correct by sending a command and checking for the correct response
        It may be that - depending on the OS - the response may be corrupted
        If this is the case try to hard-code the COM port into the config JSON file
        '''
        _available_ports = list_ports.comports(include_links=False)
        current_os = sys.platform.lower()

        # OS-specific port prefixes and descriptions
        if current_os.startswith("win"):
            ports_to_check = ["COM"]
            descriptions_to_check = ["CH340", "CP2102", "USB Serial"]
        elif current_os.startswith("darwin"):
            # prefer cu.SLAB* over others on mac
            ports_to_check = ["/dev/cu.SLAB", "/dev/cu.wchusb", "/dev/cu.usbmodem"]
            descriptions_to_check = ["CH340", "CP2102", "USB-Serial"]
        else:  # linux or other
            ports_to_check = ["/dev/ttyUSB", "/dev/ttyACM"]
            descriptions_to_check = ["CH340", "CP2102", "USB2.0-Serial", "USB-Serial"]

        for port in _available_ports:
            if any(port.device.startswith(p) for p in ports_to_check) or \
            any(port.description.startswith(d) for d in descriptions_to_check):
                if current_os.startswith("darwin") and port.device.startswith("/dev/cu.usbserial-"):
                    continue
                if self.tryToConnect(port):
                    self.is_connected = True
                    self.manufacturer = port.manufacturer
                    return self.serialdevice

        self.is_connected = False
        self.serialport = "NotConnected"
        self.serialdevice = None
        self._logger.debug("No USB device connected! Using DUMMY!")
        self.manufacturer = "UC2Mock"
        return None

    def tryToConnect(self, port):
        try:
            self.serialdevice = serial.Serial(port.device, baudrate=self.baudrate, timeout=self.read_timeout, write_timeout=self.write_timeout)
            # close the device - similar to hard reset
            if not port.description == "USB JTAG/serial debug unit": # ESP32S3 won't work like that -> non configured device afterwards 
                self.serialdevice.setDTR(False)
                self.serialdevice.setRTS(True)
                time.sleep(.1)
                self.serialdevice.setDTR(False)
                self.serialdevice.setRTS(False)
                time.sleep(.5)            
            #time.sleep(T_SERIAL_WARMUP)
            self._freeSerialBuffer(self.serialdevice, timeout=2, timeMinimum=1)
            if self.skipFirmwareCheck or self.checkFirmware(self.serialdevice):
                self.is_connected = True
                self.NumberRetryReconnect = 0
                return True
            else:
                False

        except Exception as e:
            self._logger.debug(f"Trying out port {port} failed: "+str(e))

        return False
    
    def _write(self, serialdevice, payload):
        if type(payload) == dict:
            payload = json.dumps(payload)
        serialdevice.write(payload.encode('utf-8'))
        if self.cmdWriteCallBackFct is not None:
            self.cmdWriteCallBackFct(payload)

            
    def _read(self, serialdevice):
        try:
            mLine = serialdevice.readline()
            if self.cmdReadCallBackFct is not None and mLine!=b'' and mLine != b'\n' :
                self.cmdReadCallBackFct(mLine)  
            return mLine
        except SerialException as e:
            self._logger.error("Failed to read the line in serial: "+str(e))
            return False
    
    def setWriteCallback(self, callback):
        '''
        We can assign a callback function to output any of the 
        commands that are sent to the serial device
        '''
        self.cmdWriteCallBackFct = callback
        
    def setReadCallback(self, callback):    
        '''
        We can assign a callback function to output any of the
        responses that are received from the serial device
        '''
        self.cmdReadCallBackFct = callback
    
    def checkFirmware(self, ser):
        """Check if the firmware is correct
        We do not do that inside the queue processor yet
        """
        path = "/state_get"
        payload = {"task": path}

        # write message to the serial
        self._write(ser, payload)
        ser.write(b'\n')

        #self._write(ser, b'\n')
        
        # iterate a few times in case the debug mode on the ESP32 is turned on and it sends additional lines
        for i in range(500):
            # if we just want to send but not even wait for a response
            mReadline = self._read(ser)
            if self.DEBUG and mReadline != "" and mReadline != "\n" and mReadline != b'' and mReadline != b'\n': 
                self._logger.debug("[checkFirmware]: "+str(mReadline))
            if mReadline.decode('utf-8').strip() == "++" or mReadline.decode('utf-8').strip().find("++")>=0:
                self._freeSerialBuffer(ser)
                return True
        return False

    def _generate_identifier(self):
        self.identifier_counter += 1
        return self.identifier_counter

    def _enqueue_command(self, json_command):
        """Add a command to the queue."""
        self.sender_queue.put(json_command)

    def _sending_commands(self):
        """Sending commands from the queue with a 1s delay between them."""
        while self.running:
            
            # Wait for the next command
            json_command = self.sender_queue.get()
            
            # Process the command
            with self.serialLock:
                try:
                    if self.DEBUG and json_command!="": self._logger.debug("[SendingCommands]:"+str(json_command))
                    self._write(self.serialdevice, json.loads(json_command))
                except Exception as e:
                    self._logger.error("Failed to write the line in serial: "+str(e))
            
            # Signal that the command has been processed
            self.sender_queue.task_done()
            if self.DEBUG: self._logger.debug("[SendingCommands]: Task done")
                        
            # Wait for .05 second before processing the next command
            time.sleep(0.05)
            
    def _process_commands(self):
        buffer = ""
        reading_json = False
        currentIdentifier = None
        nLineCountTimeout = 50 # maximum number of lines read before timeout
        lineCounter = 0
        nFailedCommands = 0
        nFailedCommandsMax = 10
        while self.running:
            if self.manufacturer == "UC2Mock": 
                self.running = False
                return

            # device not ready yet
            if self.serialdevice is None:
                self.is_connected = False
                continue
            else:
                self.is_connected = True

            # if we just want to send but not even wait for a response
            with self.serialLock:
                try:
                    mReadline = self._read(self.serialdevice)
                    if mReadline == False :
                        nFailedCommands += 1
                        if nFailedCommands > nFailedCommandsMax:
                            raise Exception("Failed to read the line in serial: "+str(mReadline))
                    line = mReadline.decode('utf-8').strip()
                    if self.DEBUG and line!="": 
                        self._logger.debug("[ProcessLines]:"+str(line))
                except Exception as e:
                    self._logger.error("Failed to read the line in serial: "+str(e))
                    nFailedCommands += 1
                    line = ""
                        
                    # if we have a problem with the serial connection, we need to reconnect
                    for i in range(4):
                        nFailedCommands=0
                        if self.reconnect():
                            self._logger.debug("Reconnected to the serial device")
                            break
                        else:
                            self._logger.debug("Failed to reconnect to the serial device")
                        time.sleep(1)
                
            if line == "++":
                reading_json = True
                continue
            elif line.find("error") != -1:
                # if we have an error, we need to reset the last command
                self._logger.debug("Error - last command did not match the firmware: "+str(self.commands[currentIdentifier]))
                self.resetLastCommand = True
                buffer = ""
                lineCounter = 0
                reading_json = False
                self.responses[currentIdentifier].append({"error": 1})
                self.responses[currentIdentifier].append({"qid": currentIdentifier})
            elif line == "--" or lineCounter>nLineCountTimeout:
                lineCounter = 0
                reading_json = False
                try:
                    json_response = json.loads(buffer)
                    self._logger.debug("[ProcessCommands]: "+str(json_response))
                    if len(self.callBackList) > 0:
                        for callback in self.callBackList:
                            # check if json has key
                            try:
                                if callback["pattern"] in json_response:
                                    callback["callbackfct"](json_response)
                            except Exception as e:
                                self._logger.error("[ProcessCommands Casting Callbacks]: "+str(e))

                except Exception as e:
                    self._logger.error("Failed to load the json from serial %s" % buffer)
                    self._logger.error("Error: %s" % str(e))
                    json_response = {}
                    #reading_json = True

                try: currentIdentifier = json_response["qid"]
                except: pass

                with self.lock:
                    try:
                        self.responses[currentIdentifier].append(json_response.copy())
                    except:
                        self.responses[currentIdentifier] = list()
                        self.responses[currentIdentifier].append(json_response.copy())
                    buffer = ""      # reset buffer

            # detect a reboot of the device and return the current QIDs
            elif line == "reboot":
                self._logger.warning("Device rebooted")
                self.resetLastCommand = True
                buffer = ""
                lineCounter = 0
                reading_json = False
                self.responses[currentIdentifier].append({"reboot": 1})
                self.responses[currentIdentifier].append({"qid": currentIdentifier})

            if reading_json:
                buffer += line
                lineCounter +=1
        self.running = False

    def get_json(self, path, timeout=1):
        message = {"task":path}
        message = json.dumps(message)
        return self.sendMessage(message, nResponses=0, timeout=timeout)

    def post_json(self, path, payload, getReturn=True, nResponses=1, timeout=100):
        """Make an HTTP POST request and return the JSON response"""
        if payload is None:
            payload = {}
        if "task" not in payload:
            payload["task"] = path

        # write message to the serial
        if not getReturn:
            nResponses = -1
        writeResult = self.sendMessage(command=payload, nResponses=nResponses, timeout=timeout)
        return writeResult

    def writeSerial(self, payload):
        return self.sendMessage(payload, nResponses=-1)

    def breakCurrentCommunication(self):
        pass # not needed anymore

    def readSerial(self, qid=0, timeout=1):
        t0 = time.time()
        while time.time()-t0<timeout:
            try:
                return self.responses[qid]
            except:
                pass
        return {"timeout": 1}

    def register_callback(self, callback, pattern):
        '''
        we need to add a callback function to a list of callbacks that will be read during the serial communication
        loop
        '''
        self.callBackList.append({"callbackfct":callback, "pattern":pattern})

    def sendMessage(self, command, nResponses=1, timeout = 20):
        '''
        Sends a command to the device and optionally waits for a response.
        If nResponses is 0, then the command is sent but no response is expected.
        If nResponses is 1, then the command is sent and the response is returned.
        If nResponses is >1, then the command is sent and a list of responses is returned.
        '''
        if type(command) == str:
            command = json.loads(command)
        identifier = self._generate_identifier()
        command["qid"] = identifier
        try:
            json_command = json.dumps(command)+"\n"
            #self.serialdevice.flush()
            if self.DEBUG and json_command!="": self._logger.debug("[SendingCommands]:"+str(json_command))
            with self.serialLock:
                self._write(self.serialdevice, json_command)
            #time.sleep(1)
            # we have to queue the commands and give it some time to process
            #self._enqueue_command(json_command)
            
        except Exception as e:
            if self.DEBUG: self._logger.error(e)
            return "Failed to Send"
        self.commands[identifier]=command # FIXME: Need to clear this after the response is received
        if nResponses <= 0 or not self.is_connected or self.manufacturer=="UC2Mock":
            time.sleep(0.1)
            return identifier
        t0 = time.time()
        maxRetry = 3
        iRetry = 0
        while self.running:
            time.sleep(0.002)
            if self.resetLastCommand or time.time()-t0>timeout or not self.is_connected:
                self.resetLastCommand = False
                return "communication interrupted by timeout or reset: "+str(identifier) + " and code:"+str(self.commands[identifier])
            with self.lock:
                if identifier in self.responses:
                    if len(self.responses[identifier])==nResponses:
                        returnMessage = self.responses[identifier]
                        # remove the response from the list
                        del self.responses[identifier]
                        return returnMessage
                if -identifier in self.responses:
                    self._logger.debug("You have sent the wrong command!")
                    return "Wrong Command"
            if time.time()-t0>self.timeReturnReceived and not (identifier in self.responses and len(self.responses[identifier]) > 0):
                if iRetry > maxRetry:
                    self.resetLastCommand = True
                    return "No response received"
                self._logger.debug("It takes too long to get a response, we will resend the last command: "+str(self.commands[identifier]))
                try:
                    ERROR="We have a queue, so after a while we need to resend the wrong command!"
                    iRetry += 1
                    raise Exception(ERROR)
                    self.serialdevice.write(json.dumps(self.commands[identifier]).encode('utf-8'))
                    t0 = time.time()
                    time.sleep(0.1)
                except Exception as e:
                    self._logger.error("Failed to write the line in serial: "+str(e))
            

    def interruptCurrentSerialCommunication(self):
        self.resetLastCommand = True

    def stop(self):
        self.running = False
        self.thread.join()
        self.serialdevice.close()

    def closeSerial(self):
        self.stop()
        self.running = False

    def close(self):
        self.closeSerial()
        
    def reconnect(self, baudrate=None):
        self._logger.debug("Reconnecting to the serial device")
        self.running = False
        if baudrate is not None:
            self.baudrate = baudrate
        try:
            self.serialdevice.close()
        except:
            pass
        self.serialdevice = self.openDevice(port = self.serialport, baud_rate = self.baudrate)
        if self.serialdevice: return True
        return False


if __name__ == "__main__":
    # Usage example
    monitor = Serial('/dev/cu.SLAB_USBtoUART', baudsrate=115200)  # Change to your port

    command_to_send = {
            "task": "/state_get"
    }

    command_to_send = {"task":"/motor_act","motor":{"steppers": [{ "stepperid": 3, "position": -1000, "speed": 15000, "isabs": 0, "isaccel":0}]}}

    t0 = time.time()
    response = monitor.sendMessage(command_to_send, nResponses=2)
    print("Response:", response)
    print("time:", time.time()-t0)

    t0 = time.time()
    response = monitor.sendMessage(command_to_send, nResponses=0)
    print("Response:", response)
    print("time:", time.time()-t0)


    #response = monitor.waitForResponse(command_id)

    if response:
        print("Response:", response)

    monitor.stop()


class SerialManagerWrapper:

    def __init__(self) -> None:
        pass

import random
from threading import Thread
class MockSerial:
    def __init__(self, port, baudrate, timeout=1):
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self.is_open = False
        self.data_buffer = []
        self.thread = Thread(target=self._simulate_data)
        self.thread.daemon = True
        self.thread.start()
        self.is_open = True
        self.manufacturer = "UC2Mock"
        self.BAUDRATES = -1

    def flush(self):
        pass

    def isOpen(self):
        return self.is_open

    def open(self):
        self.is_open = True

    def close(self):
        self.is_open = False

    def readline(self, timeout=1):
        if not self.is_open:
            raise Exception("Device not connected")
        if len(self.data_buffer) == 0:
            return b''
        data = self.data_buffer
        self.data_buffer = self.data_buffer
        time.sleep(.05)
        return bytes(data)

    def read(self, num_bytes):
        if not self.is_open:
            raise Exception("Device not connected")
        if len(self.data_buffer) == 0:
            return b''
        data = self.data_buffer[:num_bytes]
        self.data_buffer = self.data_buffer[num_bytes:]
        return bytes(data)

    def write(self, data):
        if not self.is_open:
            raise Exception("Device not connected")
        pass  # Do nothing, as it's a mock

    def _simulate_data(self):
        while self.is_open:
            if random.random() < 0.2:  # Simulate occasional data availability
                self.data_buffer.extend([random.randint(0, 255) for _ in range(10)])
            time.sleep(0.1)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


if __name__ == "__main__":
    # Usage example
    monitor = Serial('/dev/cu.SLAB_USBtoUART', baudrate=128000, DEBUG=True)# 115200)  # Change to your port

    command_to_send = {
            "task": "/state_get"
    }

    command_to_send = {"task":"/motor_act","qid":0,"motor":{"steppers": [{ "stepperid": 3, "position": -1000, "speed": 15000, "isabs": 0, "isaccel":0}]}}

    t0 = time.time()
    response = monitor.sendMessage(command_to_send, nResponses=2)
    print("Response:", response)
    print("time:", time.time()-t0)

    t0 = time.time()
    response = monitor.sendMessage(command_to_send, nResponses=0)
    print("Response:", response)
    print("time:", time.time()-t0)


    #response = monitor.waitForResponse(command_id)

    if response:
        print("Response:", response)


import time
import json

class Gripper(object):
    """
    Example Python class for controlling a gripper servo via JSON commands
    to the MCU, similar in style to the Objective class provided.
    """

    def __init__(self, parent):
        """
        parent: an object that handles the low-level post_json(path, payload, getReturn, timeout, nResponses)
                similar to the 'Home' or 'Objective' parent in the template.
        """
        self._parent = parent
        self.timeout = 10  # default timeout for blocking commands
        # Default angles
        self.close_angle = 0
        self.open_angle  = 180

    def close(self, isBlocking=False):
        """
        Closes the gripper to close_angle (0° by default).
        Sends {"task":"/gripper_act","action":"close"} over Serial/JSON.
        """
        path = "/gripper_act"
        payload = {
            "task": path,
            "action": "close"
        }
        # Possibly wait for the MCU response
        nResponses = 1 if isBlocking else 0
        r = self._parent.post_json(
            path,
            payload,
            getReturn=isBlocking,
            timeout=self.timeout if isBlocking else 0,
            nResponses=nResponses
        )
        return r

    def open(self, isBlocking=False):
        """
        Opens the gripper to open_angle (180° by default).
        Sends {"task":"/gripper_act","action":"open"}.
        """
        path = "/gripper_act"
        payload = {
            "task": path,
            "action": "open"
        }
        nResponses = 1 if isBlocking else 0
        r = self._parent.post_json(
            path,
            payload,
            getReturn=isBlocking,
            timeout=self.timeout if isBlocking else 0,
            nResponses=nResponses
        )
        return r

    def setAngle(self, angle, isBlocking=False):
        """
        Moves the gripper servo to an arbitrary angle (0–180).
        Sends {"task":"/gripper_act","action":"degree","value":<angle>}.
        """
        path = "/gripper_act"
        payload = {
            "task": path,
            "action": "degree",
            "value": angle
        }
        nResponses = 1 if isBlocking else 0
        r = self._parent.post_json(
            path,
            payload,
            getReturn=isBlocking,
            timeout=self.timeout if isBlocking else 0,
            nResponses=nResponses
        )
        return r

    def getstatus(self):
        """
        Query the current gripper status.
        Expects MCU to return something like:
          {
            "gripper":{
              "pos": 90,
              "isOpen": 0 or 1,
              "state": 1,
              "isRunning": 0
            }
          }
        """
        path = "/gripper_get"
        payload = {"task": path}
        r = self._parent.post_json(path, payload, timeout=1)

        try:
            # The last item in r is assumed to contain the "gripper" field
            status = r[-1]["gripper"]
        except:
            # Fallback if parsing fails
            status = {"pos": 0, "isOpen": None, "state": 0, "isRunning": 0}

        return status



import random
from threading import Thread
import time

class MockSerial:
    def __init__(self, port, baudrate, timeout=1):
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self.is_open = False
        self.data_buffer = []
        self.thread = Thread(target=self._simulate_data)
        self.thread.daemon = True
        self.thread.start()
        self.is_open = True
        self.manufacturer = "UC2Mock"
        self.BAUDRATES = -1

    def flush(self):
        pass

    def isOpen(self):
        return self.is_open

    def open(self):
        self.is_open = True

    def close(self):
        self.is_open = False

    def readline(self, timeout=1):
        if not self.is_open:
            raise Exception("Device not connected")
        if len(self.data_buffer) == 0:
            return b''
        data = self.data_buffer
        self.data_buffer = self.data_buffer
        time.sleep(.05)
        return bytes(data)

    def read(self, num_bytes):
        if not self.is_open:
            raise Exception("Device not connected")
        if len(self.data_buffer) == 0:
            return b''
        data = self.data_buffer[:num_bytes]
        self.data_buffer = self.data_buffer[num_bytes:]
        return bytes(data)

    def write(self, data):
        if not self.is_open:
            raise Exception("Device not connected")
        pass  # Do nothing, as it's a mock

    def _simulate_data(self):
        while self.is_open:
            if random.random() < 0.2:  # Simulate occasional data availability
                self.data_buffer.extend([random.randint(0, 255) for _ in range(10)])
            time.sleep(0.1)

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()


if __name__ == '__main__':
    with MockSerial("COM1", 9600) as ser:
        while True:
            data = ser.readline()
            print(data)
            time.sleep(1)


import numpy as np
import time
import json


gTIMEOUT = 1 # seconds to wait for a response from the ESP32
class Message(object):
    ''' 
    This class only parses incoming messages from the ESP32 that can be used e.g. for triggering events such as taking an image
    or converting hardware inputs such as button presses to software events
    '''
    def __init__(self, parent=None, nCallbacks = 10):
        self._parent = parent
        self.nCallbacks = nCallbacks
        # initialize the callback functions
        self.init_callback_functions(self.nCallbacks)
        
        # register a callback function for the motor status on the serial loop
        if hasattr(self._parent, "serial"):
            self._parent.serial.register_callback(self._callback_message, pattern="message")


    def _callback_message(self, data):
        ''' cast the json in the form:
        {
            "message":      {
                    "key":  1,
                    "data": 1
            }
        }       
        into seperated actions that follow this pattern:
        data lookup:
		{
			key 	|	 meaning 		|	value
			--------------------------------
			1		|	snap image		|	0
			2		|	exposure time	|	0....1000000
			3		|	gain			|	0...100
            ...
            
        Example:
            #%%
            import uc2rest
            import time
            
            port = "unknown"
            port = "/dev/cu.SLAB_USBtoUART"
            ESP32 = uc2rest.UC2Client(serialport=port, baudrate=500000, DEBUG=True)

            # register callback function for key/value pair 
            def my_callback_key1(value):
                print("Callback: ", value)
            ESP32.message.register_callback(1, my_callback_key1)

            while True:
                time.sleep(.1)
		*/
  
        '''
        try:
            # parse the message
            key = data["message"]["key"]
            value = data["message"]["data"]
            # trigger the action
            if self._callbackPerKey[key] is not None:
                self._callbackPerKey[key](value) # we call the function with the value
        except Exception as e:
            print("Error in _callback_message: ", e)

    def init_callback_functions(self, nCallbacks=10):
        ''' initialize the callback functions '''
        self._callbackPerKey = {}
        self.nCallbacks = nCallbacks
        for i in range(nCallbacks):
            self._callbackPerKey[i] = None
            
    def register_callback(self, key, callback):
        ''' register a callback function for a specific key '''
        self._callbackPerKey[key] = callback
        
    def trigger_message(self, key:int=1, value:int=1):
        # {"task": "/message_act", "message": 1, "key":1, "value":1}
        path = "/message_act"
        payload = {
            "task": path,
            "key": key,
            "value": value
            }
        r = self._parent.post_json(path, payload, getReturn=False)
        return r



import time
import threading
class Temperature(object):

    def __init__(self, parent):
        self._parent = parent
        self._temperature = -273.15
        
        self.is_temperature_polling = False
        
        # register a callback function for the motor status on the serial loop
        if hasattr(self._parent, "serial"):
            self._parent.serial.register_callback(self._callback_temperature_status, pattern="heat")
        

    '''
    ##############################################################################################################################
    Temperature Controllers
    ##############################################################################################################################
    '''
    
    def start_temperature_polling(self, period=5):
        self.is_temperature_polling = True
        def _poll_temperature():
            while self.is_temperature_polling:
                path = "/heat_get"
                r = self._parent.get_json(path, timeout=0)
                time.sleep(period)
        threading.Thread(target=_poll_temperature).start()
        
    def stop_temperature_polling(self):
        self.is_temperature_polling = False
        
        
    def _callback_temperature_status(self, data):
        ''' cast the json in the form:
        {
        "qid":	0,
        "heat":	37.0
        }
        into the position array of the motors '''
        try:
            self._temperature = data["heat"]
        except Exception as e:
            print("Error in _callback_temperature_status: ", e)


    def stop_heating(self):
        r = self.set_temperature(active=0)
        return r

    def set_temperature(self, active=1, Kp=1000, Ki=0.1, Kd=0.1, target=37, timeout=600000000, updaterate=1000):
        # {"task": "/heat_act", "active":1, "Kp":1000, "Ki":0.1, "Kd":0.1, "target":37, "timeout":60000000, "updaterate":1000}
        path = "/heat_act"
        payload = {
            "task": path,
            "active": int(active),
            "Kp": Kp,
            "Ki": Ki,
            "Kd": Kd,
            "target": target,
            "timeout": timeout,
            "updaterate": updaterate
            }
        r = self._parent.post_json(path, payload, getReturn=False)
        return r

    def get_temperature(self, timeout=0.5, isBlocking=False):
        # {"task": "/heat_get"}
        if isBlocking:
            path = "/heat_get"
            r = self._parent.get_json(path, timeout=timeout)
            try:
                r = r["heat"]
            except:
                r = -273.15
            return r
        else:
            return self._temperature

if __name__ == "__main__":

    #%%
    import uc2rest
    import time
    port = "/dev/cu.SLAB_USBtoUART"
    ESP32 = uc2rest.UC2Client(serialport=port, DEBUG=True)
    ESP32.temperature.set_temperature(active=1, Kp=1000, Ki=0.1, Kd=0.1, target=37, timeout=600000, updaterate=1000)
    for i in range(10):
        print(ESP32.temperature.get_temperature())
        time.sleep(1)
    ESP32.temperature.stop_heating()
    ESP32.close()



import time
import json

class Objective(object):
    def __init__(self, parent):
        """
        parent: an object that handles the low-level post_json(path, payload, getReturn, timeout, nResponses)
        similar to the 'Home' class example.
        """
        self._parent = parent
        # Default parameters (adjust as needed)
        self.speed = 20000
        self.accel = 20000
        self.direction = -1
        self.endstoppolarity = 1
        self.timeout = 10
        # Default objective positions
        self.x1 = 1000
        self.x2 = 2000
        self.z1 = 0
        self.z2 = 0
        self.objectiveAxis = 0

        
    def home(self, direction=None, endstoppolarity=None, speed=None, acceleration=None, hometimeout=20000, axis=None, isBlocking=False):
        # {"task":"/home_act", "home": {"steppers": [{"stepperid":0, "timeout": 20000, "speed": 10000, "direction":-1, "endstoppolarity":1}]}}
        if axis is None:
            axis = self.objectiveAxis
        if endstoppolarity is None:
            endstoppolarity = self.endstoppolarity
        if direction is None:
            direction = self.direction
        if speed is None:
            speed = self.speed
        if acceleration is None:
            acceleration = self.accel
        path = "/home_act"
        payload = {
            "task": path,
            "home": {
                "steppers": [
                    {
                        "stepperid": axis,
                        "timeout": hometimeout,
                        "speed": speed,
                        "direction": direction,
                        "endstoppolarity": endstoppolarity, 
                    }
                ]
            }
        }
        nResponses = 2 if isBlocking else 0
        r = self._parent.post_json(path, payload,
                                        getReturn=isBlocking, 
                                        timeout=self.timeout if isBlocking else 0, 
                                        nResponses=nResponses)
        return r
        
    def setX1(self, x1):
        self.x1 = x1
        
    def setX2(self, x2):
        self.x2 = x2
        
    def setZ1(self, z1):
        self.z1 = z1
    
    def setZ2(self, z2):
        self.z2 = z2
        
    def getstatus(self):
        """
        Get the objective status.
        returns:  
        {"objective":{"x1":1000,"x2":2000,"z1":0, "z2":0, "pos":1000,"isHomed":1,"state":1,"isRunning":0}}
        """
        path = "/objective_get"
        payload = {
            "task": path,
        }
        r = self._parent.post_json(path, payload, timeout=1)
        # parse the response
        try:
            status = r[-1]["objective"]
            status["x1"] = np.round(status["x1"]*self._parent.motor.stepSizeA)
            status["x2"] = np.round(status["x2"]*self._parent.motor.stepSizeA)
            status["z1"] = np.round(status["z1"]*self._parent.motor.stepSizeZ)
            status["z2"] = np.round(status["z2"]*self._parent.motor.stepSizeZ)
        except:
            status = {"x1":0, 
                      "x2":0, 
                      "z1":0,
                      "z2":0,
                      "pos":0, 
                      "isHomed":0, 
                      "state":0, 
                      "isRunning":0}
        return status
        
    def calibrate(self, direction=None, endstoppolarity=None, isBlocking=False):
        """
        Calibrate the objective (sets the position=0 position).
        direction => homing direction (e.g. 1 or -1)
        endstoppolarity => polarity for endstop (e.g. 1 or -1)
        isBlocking => if True, waits for the MCU's response (home done); else returns immediately
        """
        if direction is None:
            direction = self.direction
        if endstoppolarity is None:
            endstoppolarity = self.endstoppolarity

        path = "/objective_act"
        payload = {
            "task": path,
            "calibrate": 1,
            "homeDirection": direction,
            "homeEndStopPolarity": endstoppolarity
        }

        # if blocking, wait for two responses: "ok" + "home done"
        nResponses = 2 if isBlocking else 0 #
        r = self._parent.post_json(
            path, 
            payload, 
            getReturn=isBlocking, 
            timeout=self.timeout if isBlocking else 0, 
            nResponses=nResponses
        )
        return r

    def toggle(self, speed=None, accel=None, isBlocking=False):
        """
        Toggle between x1 and x2.
        speed, accel => override default speeds
        """
        if speed is None:
            speed = self.speed
        if accel is None:
            accel = self.accel

        path = "/objective_act"
        payload = {
            "task": path,
            "toggle": 1,
            "speed": abs(speed),
            "accel": abs(accel)
        }

        # if blocking, wait for two responses: "ok" + "motor done"
        nResponses = 2 if isBlocking else 0
        r = self._parent.post_json(
            path, 
            payload, 
            getReturn=isBlocking, 
            timeout=self.timeout if isBlocking else 0, 
            nResponses=nResponses
        )
        return r

    def move(self, slot=1, speed=None, accel=None, isBlocking=False):
        """
        Move to slot 1 (x1/z1) or slot 2 (x2/z2).
        slot => 1 or 2
        speed, accel => override default speeds
        """
        if speed is None:
            speed = self.speed
        if accel is None:
            accel = self.accel

        path = "/objective_act"
        payload = {
            "task": path,
            "move": slot,
            "speed": abs(speed),
            "accel": abs(accel),
            "obj": slot  # optional if you want to include it exactly as shown
        }

        nResponses = 2 if isBlocking else 0
        r = self._parent.post_json(
            path, 
            payload, 
            getReturn=isBlocking, 
            timeout=self.timeout if isBlocking else 0, 
            nResponses=nResponses
        )
        return r

    def setPositions(self, x1=None, x2=None, z1=None, z2=None, isBlocking=False):
        """
        Set explicit positions in steps.
        If x1 == -1, the current position is used as x1.
        Example:
            {"task":"/objective_act", "x1": 1000, "x2": 2000}
            {"task":"/objective_act", "x1": -1, "x2": 2000}
            {"task":"/objective_act", "x1": 1000, "x2": -1, "z1": 1, "z2": 100}
        """
        if x1 is not None:
            self.x1 = x1
        if x2 is not None:
            self.x2 = x2
        if z1 is not None:
            self.z1 = z1
        if z2 is not None:
            self.z2 = z2

        path = "/objective_act"
        payload = {"task": path}

        # Only include x1 or x2 in JSON if they were explicitly provided
        # If x1 == -1, that instructs the MCU to take current motor position
        if x1 is not None:
            payload["x1"] = x1/self._parent.motor.stepSizeA
        if x2 is not None:
            payload["x2"] = x2/self._parent.motor.stepSizeA
        if z1 is not None:
            payload["z1"] = z1/self._parent.motor.stepSizeZ
        if z2 is not None:
            payload["z2"] = z2/self._parent.motor.stepSizeZ

        nResponses = 1 if isBlocking else 0
        r = self._parent.post_json(
            path, 
            payload, 
            getReturn=isBlocking, 
            timeout=self.timeout if isBlocking else 0, 
            nResponses=nResponses
        )
        return r



import numpy as np

gTimeout = 2

class LedMatrix(object):
    def __init__(self, parent, NLeds=64):
        """
        This class handles sending JSON commands for controlling a NeoPixel or LED matrix device.
        The 'parent' is assumed to be an object providing post_json(...) and get_json(...) methods.
        """
        self.NLeds = NLeds
        self.Nx = self.Ny = int(np.sqrt(NLeds)) if NLeds > 0 else 8

        # We assume the pattern is binary (0 or 1) for an NxN grid, stored in ledpattern NxN or flattened.
        # We store an RGB for each LED if needed. -1 indicates "unset" or no color assigned.
        self.ledpattern = np.ones((self.NLeds, 3)) * -1

        self._parent = parent  # must have post_json() etc.
        self.timeout = 1
        self.intensity = (255, 255, 255)

        # Maps textual modes to an integer, used for backward compatibility
        self.ledArrayModes = {
            "array": 0,
            "full": 1,
            "single": 2,
            "off": 3,
            "left": 4,
            "right": 5,
            "top": 6,
            "bottom": 7,
            "multi": 8
        }

        self.currentLedArrayMode = "full"

    ######################################################################################################
    # BASE / LEGACY METHODS
    ######################################################################################################

    def send_LEDMatrix_array(self, led_pattern, getReturn=True, timeout=gTimeout):
        """
        Send an LED array pattern e.g. a list of { id, r, g, b } or a NumPy array shape (N,3).
        Produces JSON:
        {
          "task": "/ledarr_act",
          "led": {
            "action": "array",
            "LEDArrMode": 0,
            "led_array": [
               {"id":0,"r":...,"g":...,"b":...},
               ...
            ]
          }
        }
        """
        path = "/ledarr_act"

        # If led_pattern is a NumPy array, convert it to a list of dict
        if not isinstance(led_pattern, list):
            if len(led_pattern.shape) == 3:
                # flatten 3D into 2D
                led_pattern = np.reshape(led_pattern, (led_pattern.shape[0]*led_pattern.shape[1], led_pattern.shape[2]))

            pattern_list = []
            for i in range(led_pattern.shape[0]):
                pattern_list.append({
                    "id": i,
                    "r": int(led_pattern[i, 0]),
                    "g": int(led_pattern[i, 1]),
                    "b": int(led_pattern[i, 2])
                })
        else:
            pattern_list = led_pattern

        payload = {
            "task": path,
            "qid": 0,  # or fill in dynamically if you like
            "led": {
                "action": "array",
                "LEDArrMode": self.ledArrayModes["array"],
                "led_array": pattern_list
            }
        }

        r = self._parent.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        if not getReturn or timeout == 0:
            r = {"success": 1}
        self.currentLedArrayMode = "array"
        return r

    def send_LEDMatrix_full(self, intensity=(255, 255, 255), getReturn=True, timeout=gTimeout):
        """
        Fill all LEDs with the same (r,g,b). 
        Creates JSON with: "action": "fill"
        """
        path = "/ledarr_act"
        payload = {
            "task": path,
            "qid": 0,
            "led": {
                "action": "fill",
                "LEDArrMode": self.ledArrayModes["full"],
                "r": int(intensity[0]),
                "g": int(intensity[1]),
                "b": int(intensity[2])
            }
        }

        r = self._parent.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        if not getReturn or timeout == 0:
            r = {"success": 1}
        self.currentLedArrayMode = "full"
        return r

    def send_LEDMatrix_single(self, indexled=0, intensity=(255, 255, 255), getReturn=True, timeout=gTimeout):
        """
        Update only a single LED with color (r,g,b).
        Creates JSON: "action":"single"
        """
        path = "/ledarr_act"
        payload = {
            "task": path,
            "qid": 0,
            "led": {
                "action": "single",
                "LEDArrMode": self.ledArrayModes["single"],
                "ledIndex": int(indexled),
                "r": int(intensity[0]),
                "g": int(intensity[1]),
                "b": int(intensity[2])
            }
        }

        r = self._parent.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        if not getReturn or timeout == 0:
            r = {"success": 1}
        self.currentLedArrayMode = "single"
        return r

    def get_LEDMatrix(self, timeout=gTimeout):
        """
        Request info from the device about pin/LED count, etc.
        Usually the device responds with JSON containing these details.
        """
        path = "/ledarr_get"
        payload = {"task": path}
        r = self._parent.post_json(path, payload, getReturn=True, timeout=timeout)
        return r

    def setSingle(self, indexled, state):
        """
        High-level usage: set a single LED's "state" (0 or 1), then send it with intensity scaling.
        """
        ix = indexled // self.Nx
        iy = indexled % self.Nx
        self.ledpattern[indexled] = (state, state, state)
        # If your wiring is zig-zag or reversed, you may need to adapt index
        if ix % 2 != 0:
            indexled = (ix * self.Nx) + (self.Ny - iy - 1)

        intensityScaled = np.array(state) * np.array(self.intensity)
        self.send_LEDMatrix_single(indexled=indexled, intensity=intensityScaled, timeout=self.timeout)
        return self.ledpattern

    def setAll(self, state: int, intensity:int =None, getReturn=True):
        """
        Turn on or off all LEDs at a certain intensity. 
        If state is a boolean array or a single boolean, we set them all.
        """
        onval = np.sum(state) > 0
        if intensity is not None:
            self.intensity = intensity
        intensity2display = np.array(self.intensity) * onval
        self.send_LEDMatrix_full(intensity=intensity2display, getReturn=getReturn)
        self.ledpattern = onval * np.ones((self.NLeds, 3))
        return self.ledpattern

    def setIntensity(self, intensity:int, getReturn:bool=True):
        self.intensity = intensity
        self.setPattern(getReturn=getReturn)

    def setPattern(self, getReturn:bool=True, ledpattern=None):
        """
        If ledpattern is provided, store it, then send to device.
        If the entire pattern is "on", we do a 'full' fill for speed,
        otherwise we do 'array' updates for each LED.
        """
        if ledpattern is not None:
            self.ledpattern = ledpattern
        pattern2send = (self.ledpattern >= 1) * self.intensity

        if np.sum(self.ledpattern, 0)[0] == self.ledpattern.shape[0]:
            # All on => just do a fill
            self.send_LEDMatrix_full(pattern2send[0, :], getReturn=getReturn)
        else:
            # partial => send an array of per-LED
            self.send_LEDMatrix_array(pattern2send, getReturn=getReturn)
        return self.ledpattern

    def getPattern(self):
        return self.ledpattern

    def set_led(self, colour=(0, 0, 0)):
        """
        (Legacy) Possibly sets a single color. 
        Demo only - not used in the new command style.
        """
        path = "/led"
        payload = {"r": colour[0], "g": colour[1], "b": colour[2]}
        r = self._parent.post_json(path, payload)
        return r

    def get_ledpin(self):
        """
        Another legacy method that might do a GET request.
        """
        path = "/ledarr_get"
        r = self._parent.get_json(path, getReturn=True, timeout=1)
        return r

    ######################################################################################################
    # NEW METHODS FOR "OFF", "RINGS", "CIRCLES", "HALVES", etc.
    ######################################################################################################
    def send_LEDMatrix_off(self, getReturn=True, timeout=gTimeout):
        """
        Turn all LEDs off.
        JSON: "action":"off"
        """
        path = "/ledarr_act"
        payload = {
            "task": path,
            "qid": 0,
            "led": {
                "action": "off"
            }
        }
        r = self._parent.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        return r

    def send_LEDMatrix_halves(self, region="left", intensity=(255,255,255), getReturn=True, timeout=gTimeout):
        """
        Light only "left"/"right"/"top"/"bottom" half in color, rest is dark.
        JSON: "action":"halves", "region": <string>
        """
        if type(intensity)==int:
            intensity = (intensity, intensity, intensity)        
        path = "/ledarr_act"
        payload = {
            "task": path,
            "qid": 0,
            "led": {
                "action": "halves",
                "region": region,
                "r": int(intensity[0]),
                "g": int(intensity[1]),
                "b": int(intensity[2])
            }
        }
        r = self._parent.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        return r

    def send_LEDMatrix_rings(self, radius=4, intensity=(255,255,255), getReturn=True, timeout=gTimeout):
        """
        Draw a ring of radius N. Often done by filling the circle and carving out center to make it hollow.
        JSON: "action":"rings", "radius": <int>
        """
        if type(intensity)==int:
            intensity = (intensity, intensity, intensity)
        path = "/ledarr_act"
        payload = {
            "task": path,
            "qid": 0,
            "led": {
                "action": "rings",
                "radius": radius,
                "r": int(intensity[0]),
                "g": int(intensity[1]),
                "b": int(intensity[2])
            }
        }
        r = self._parent.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        return r

    def send_LEDMatrix_circles(self, radius=4, intensity=(255,255,255), getReturn=True, timeout=gTimeout):
        """
        Draw a filled circle of radius N in color.
        JSON: "action":"circles", "radius": <int>
        """
        if type(intensity)==int:
            intensity = (intensity, intensity, intensity)
        path = "/ledarr_act"
        payload = {
            "task": path,
            "qid": 0,
            "led": {
                "action": "circles",
                "radius": radius,
                "r": int(intensity[0]),
                "g": int(intensity[1]),
                "b": int(intensity[2])
            }
        }
        r = self._parent.post_json(path, payload, getReturn=getReturn, timeout=timeout)
        return r

    def send_LEDMatrix_status(self, status="idle"):
        """
        Set the status of the LED matrix to "idle" or "busy".
        JSON: "action":"status", "status": <string>
        """
        if status not in ["error", "idle", "warn", "success", "busy", "rainbow"]:
            status = "idle"
        path = "/ledarr_act"
        payload = {
            "task": path,
            "qid": 0,
            "led": {
                "action": "status",
                "status": status
            }
        }
        r = self._parent.post_json(path, payload)
        return r


import time
import json        
        
class Home(object):
    ## Laser
    def __init__(self, parent):
        self._parent = parent
        self.endposrelease = 3000
        self.direction = 1
        self.speed = 15000
        self.timeout = 20000
        self.endstoppolarity = 1
        
    def home_x(self, speed = None, direction = None, endposrelease = None, endstoppolarity=None, timeout=None, isBlocking=False):
        # axis = 1 corresponds to 'X'
        axis = 1
        self.home(axis=axis, 
                  endstoptimeout=timeout, 
                  speed = speed, 
                  direction = direction, 
                  endposrelease=endposrelease,
                  endstoppolarity=endstoppolarity,
                  isBlocking=isBlocking)

    def home_y(self, speed = None, direction = None, endposrelease = None, endstoppolarity=None, timeout=None, isBlocking=False):
        # axis = 2 corresponds to 'Y'
        axis = 2
        self.home(axis=axis, 
                  endstoptimeout=timeout, 
                  speed = speed, 
                  direction = direction, 
                  endposrelease=endposrelease, 
                  endstoppolarity=endstoppolarity,
                  isBlocking=isBlocking)    
    
    def home_z(self, speed = None, direction = None, endposrelease = None, endstoppolarity=None, timeout=None, isBlocking=False):
        # axisa = 3 corresponds to 'Z'
        axis = 3
        self.home(axis=axis, 
                  endstoptimeout=timeout, 
                  speed = speed, 
                  direction = direction, 
                  endposrelease=endposrelease, 
                  endstoppolarity=endstoppolarity,
                  isBlocking=isBlocking)    
        
    def home_a(self, speed = None, direction = None, endposrelease = None, endstoppolarity=None, timeout=None, isBlocking=False):
        # axis = 0 corresponds to 'A'
        axis = 0
        self.home(axis=axis,
                  endstoptimeout=timeout, 
                  speed = speed, 
                  direction = direction, 
                  endposrelease=endposrelease, 
                  endstoppolarity=endstoppolarity,
                  isBlocking=isBlocking)    
    
    def home(self, axis=None, timeout=None, speed=None, direction=None, endposrelease=None, endstoppolarity=None, endstoptimeout=10000, isBlocking=False):
        '''
        axis = 0,1,2,3 or 'A, 'X','Y','Z'
        timeout => when to stop homing (it's a while loop on the MCU)
        speed => speed of homing (0...15000)
        direction => 1,-1 (left/right)
        endposrelease => how far to move after homing (0...3000)
        '''
        
        # default values
        if speed is None:
            speed = self.speed
        if direction is None:
            direction = self.direction
        if endposrelease is None:
            endposrelease = self.endposrelease
        if timeout is None:
            timeout = self.timeout
        if endstoppolarity is None:
            endstoppolarity = self.endstoppolarity

        if direction not in [-1,1]:
            direction = 1

        # construct json string
        path = "/home_act"

        payload = {
            "task": path,
            "home":{
                "steppers": [
                {
                 "stepperid": axis,
                 "timeout":endstoptimeout,
                 "speed":abs(speed),
                 "direction":direction,
                 "endstoppolarity":endstoppolarity
                 }]
            }}
     
        timeout = timeout if isBlocking else 0
        nResponses = 2 # one for command received, one for home reached
        
        # if we get a return, we will receive the latest position feedback from the driver  by means of the axis that moves the longest
        r = self._parent.post_json(path, payload, getReturn=isBlocking, timeout=timeout, nResponses=nResponses)

        return r




from datetime import datetime as dt


def format_date(dt_, fmt="%m/%d/%Y, %H:%M:%S"):
    return dt_.strftime(fmt)


def now(fmt="%m/%d/%Y, %H:%M:%S"):
    return format_date(dt.now(), fmt)


def remove_class(element, className):
    element.element.classList.remove(className)


def add_class(element, className):
    element.element.classList.add(className)
